<?php

namespace goodboyalex\php_components_pack\extensions;

use Closure;
use DateMalformedStringException;
use DateTime;
use DateTimeImmutable;
use DateTimeInterface;
use Exception;
use goodboyalex\php_components_pack\exceptions\TypeException;
use UnitEnum;

/**
 * Расширение для любого типа.
 *
 * @author Александр Бабаев
 * @package php_components_pack
 * @version 1.0
 * @since 1.1.1
 */
final class TypeExtension
{
    /**
     * Дефолтный метод обработки классов для ToArray.
     *
     * @return Closure Метод обработки классов для ToArray.
     */
    public static function DEFAULT_TO_ARRAY_ON_CLASS (): Closure
    {
        return fn (object $class) => self::ToArray($class, self::DEFAULT_TO_ARRAY_ON_CLASS());
    }

    /**
     * Переводит объект в массив.
     *
     * @param object $class Объект.
     * @param callable $onClass Метод обработки классов.
     *
     * @return array Массив свойств типа.
     * @throws TypeException Исключение, если объект не является классом.
     */
    public static function ToArray (object $class, callable $onClass): array
    {
        // Создаю массив результата
        $result = [];

        // Сперва проверяю, что объект является классом
        if (!is_object($class))
            // - если нет, то выбрасываю исключение
            throw new TypeException($class::class, "Object is not class / Объект не является классом.");

        // Затем записываю имя класса
        $result["type_class"] = $class::class;

        // Получаю свойства класса
        $properties = get_object_vars($class);

        // Для каждого свойства класса
        foreach ($properties as $key => $value) {
            // - если значение является датой
            if ($value instanceof DateTimeInterface) {

                // -- добавляю в массив
                $result[$key] = $value->getTimestamp();

                // -- следующий элемент
                continue;
            }

            // - если значение является перечислением
            if ($value instanceof UnitEnum) {
                // -- получаю перечисление
                $result[$key] = $value->name;

                // -- следующий элемент
                continue;
            }

            // - если значение является NULL
            if (is_null($value)) {
                // -- присваиваю NULL
                $result[$key] = "null";
                // -- следующий элемент
                continue;
            }

            // - если значение является классом
            if (is_object($value)) {
                // -- добавляю в массив через рекурсию
                $result[$key] = $onClass($value);

                // -- следующий элемент
                continue;
            }

            // - присваиваю значение
            $result[$key] = $value;
        }

        // Возвращаю результат
        return $result;
    }

    /**
     * Переводит массив в объект.
     *
     * @param array $array Массив свойств.
     *
     * @return mixed Объект типа $className.
     * @throws TypeException Исключение, если массив не создан через ToArray, объект не создан, дата не корректного
     *     формата или пытается передать null не nullable типу.
     */
    public function FromArray (array $array): object
    {
        // Проверяю, что массив является массивом, созданным через ToArray, то есть содержит ключ type_class
        if (!array_key_exists("type_class", $array))
            // - если нет, то выбрасываю исключение
            throw new TypeException("array", "Array is not created by ToArray / Массив не создан через ToArray.");

        // Получаю имя класса
        $className = $array["type_class"];

        // Создаем объект
        $instance = new $className();

        // Проверяем, что объект создан
        if (!is_object($instance) && $instance instanceof $className)
            // - если нет, то выбрасываю исключение
            throw new TypeException($className, "Object is not created / Объект не создан.");

        // Получаю свойства класса
        $properties = get_object_vars($instance);

        // Для каждого свойства класса
        foreach ($properties as $key => $value) {
            // - проверяю, есть ли свойство в массиве
            if (!array_key_exists($key, $array))
                // -- если нет, то пропускаю
                continue;

            // - получаю значение из массива
            $value = $array[$key];

            // - если значение является датой DateTime
            if ($instance->$key instanceof DateTime && is_int($value)) {
                // -- присваиваю значение
                try {
                    $instance->$key = new DateTime ('@' . $value);
                }
                catch (DateMalformedStringException) {
                    throw new TypeException($className . "->{$key}::DateTime",
                        "Date is not correct format / Дата не корректного формата.");
                }

                // -- следующий элемент
                continue;
            }

            // - если значение является датой DateTimeImmutable
            if ($instance->$key instanceof DateTimeImmutable && is_int($value)) {
                // -- присваиваю значение
                try {
                    $instance->$key = new DateTimeImmutable('@' . $value);
                }
                catch (DateMalformedStringException) {
                    throw new TypeException($className . "->{$key}::DateTimeImmutable",
                        "Date is not correct format / Дата не корректного формата.");
                }

                // -- следующий элемент
                continue;
            }

            // - если значение является перечислением
            if ($instance->$key instanceof UnitEnum) {
                // -- получаю все значения перечисления
                foreach ($instance->$key::getCases() as $case)
                    // --- если значение совпадает
                    if ($case->name === $value)
                        // ---- присваиваю значение
                        $instance->$key = $case;

                // -- следующий элемент
                continue;
            }

            // - если значение является NULL
            if ($value = "null") {
                // -- присваиваю NULL
                try {
                    $instance->$key = null;
                }
                catch (Exception $exception) {
                    throw new TypeException($className . "->$key",
                        "Cannot assign null to non-nullable property. / Невозможно присвоить null не nullable типу: {$exception->getMessage()}.");
                }

                // -- следующий элемент
                continue;
            }

            // - если значение является классом
            if (is_object($instance->$key) && is_array($array[$key]) && array_key_exists("type_class", $array[$key])) {
                // -- добавляю в массив через рекурсию
                $instance->$key = self::FromArray($array[$key]);

                // -- следующий элемент
                continue;
            }

            // - присваиваю значение
            $instance->$key = $value;
        }

        // Возвращаю объект
        return $instance;
    }
}